@@ -31,11 +31,7 @@ setGeneric("elementNROWS", function(x) standardGeneric("elementNROWS"))
 setMethod("elementNROWS", "ANY", sapply_NROW)
 
 ### Used in the SGSeq package!
-quick_togroup <- function(x)
-{
-    x_eltNROWS <- elementNROWS(x)
-    rep.int(seq_along(x_eltNROWS), x_eltNROWS)
-}
+quick_togroup <- function(x) map_inner_ROWS_to_list_elements(elementNROWS(x))
 
 setMethod("elementNROWS", "List",
     function(x)
