@@ -450,24 +450,44 @@ setReplaceMethod("[", "Vector",
     {
         if (!missing(j) || length(list(...)) > 0L)
             stop("invalid subsetting")
-        i <- normalizeSingleBracketSubscript(i, x, as.NSBS=TRUE)
-        li <- length(i)
+        nsbs <- normalizeSingleBracketSubscript(i, x, as.NSBS=TRUE,
+                                                allow.append=TRUE)
+        li <- length(nsbs)
         if (li == 0L) {
             ## Surprisingly, in that case, `[<-` on standard vectors does not
             ## even look at 'value'. So neither do we...
             return(x)
         }
-        lv <- NROW(value)
-        if (lv == 0L)
-            stop("replacement has length zero")
         value <- normalizeSingleBracketReplacementValue(value, x)
-        if (li != lv) {
-            if (li %% lv != 0L)
-                warning("number of values supplied is not a sub-multiple ",
-                        "of the number of values to be replaced")
-            value <- extractROWS(value, rep(seq_len(lv), length.out=li))
+        if (is.null(value)) {
+            return(extractROWS(x, complement(nsbs)))
         }
-        replaceROWS(x, i, value)
+        value <- recycleSingleBracketReplacementValue(value, x, nsbs)
+        mergeROWS(x, i, value)
+    }
+)
+
+setMethod("mergeROWS", c("Vector", "ANY"),
+    function(x, i, value)
+    {
+        nsbs <- normalizeSingleBracketSubscript(i, x, as.NSBS=TRUE,
+                                                allow.append=TRUE)
+        if (max(nsbs) <= NROW(x)) {
+            nsbs@upper_bound_is_strict <- TRUE
+            return(replaceROWS(x, nsbs, value))
+        }
+        idx <- as.integer(nsbs)
+        oob <- idx > NROW(x)
+        value_idx <- integer(max(nsbs) - NROW(x))
+        ## handles replacement in the appended region
+        value_idx[idx[oob] - NROW(x)] <- seq_along(value)[oob]
+        if (any(value_idx == 0L)) {
+            stop("appending gaps is not supported")
+        }
+        new_values <- extractROWS(value, value_idx)
+        names(new_values) <- if (is.character(i)) i[oob] else NULL
+        x <- bindROWS(x, list(new_values), check=FALSE)
+        replaceROWS(x, idx[!oob], extractROWS(value, !oob))
     }
 )
 
@@ -513,7 +533,6 @@ setMethod("replaceROWS", c("Vector", "ANY"),
     }
 )
 
-
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Convenience wrappers for common subsetting operations
 ###
@@ -596,7 +615,7 @@ rbind_mcols <- function(...)
     all_colnames <- unique(unlist(colnames_list, use.names=FALSE))
     fillCols <- function(df) {
         if (nrow(df))
-            df[setdiff(all_colnames, colnames(df))] <- DataFrame(NA)
+            df[setdiff(all_colnames, colnames(df))] <- NA
         df
     }
     do.call(rbind, lapply(mcols_list, fillCols))
