@@ -110,6 +110,16 @@ setMethod("isStrictlySorted", "NSBS",
     function(x) isStrictlySorted(as.integer(x))
 )
 
+setMethod("max", "NSBS", function (x, ..., na.rm = FALSE) {
+    max(x@subscript, ..., na.rm=na.rm)
+})
+
+setGeneric("complement", function(x) standardGeneric("complement"))
+
+setMethod("complement", "NSBS", function(x) {
+    subscript <- which(tabulate(as.integer(x), x@upper_bound) == 0L)
+    NativeNSBS(subscript, x@upper_bound, TRUE, FALSE)
+})
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### NativeNSBS objects.
@@ -399,7 +409,7 @@ normalizeSingleBracketSubscript <- function(i, x, exact=TRUE,
 
 ### Dispatch on the 2nd argument!
 setGeneric("normalizeSingleBracketReplacementValue", signature="x",
-    function(value, x, i)
+    function(value, x)
         standardGeneric("normalizeSingleBracketReplacementValue")
 )
 
@@ -425,12 +435,44 @@ setMethod("normalizeSingleBracketReplacementValue", "ANY",
     }
 )
 
+setMethod("normalizeSingleBracketReplacementValue", "List",
+          function(value, x)
+          {
+              if (is.null(value))
+                  return(NULL)
+              callNextMethod()
+          }
+)
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### extractROWS(), replaceROWS()
+### recycleSingleBracketReplacementValue()
 ###
-### 2 internal generics to ease implementation of [ and [<- subsetting for
-### Vector subclasses.
+
+recycleSingleBracketReplacementValue <- function(value, x, i) {
+    if (is.null(value))
+        return(NULL)
+    i <- normalizeSingleBracketSubscript(i, x, allow.append=TRUE, as.NSBS=TRUE)
+    li <- length(i)
+    if (li == 0L)
+        return(value)
+    lv <- NROW(value)
+    if (lv == 0L)
+        stop("replacement has length zero")
+    if (li != lv) {
+        if (li %% lv != 0L)
+            warning("number of values supplied is not a sub-multiple ",
+                    "of the number of values to be replaced")
+        value <- extractROWS(value, rep(seq_len(lv), length.out=li))
+    }
+    value
+}
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### extractROWS(), replaceROWS(), mergeROWS(), extractCOLS(), replaceCOLS()
+###
+### 5 internal generics to ease implementation of [ and [<- subsetting for
+### Vector and DataFrame subclasses.
 ###
 ### A Vector subclass Foo should only need to implement an "extractROWS" and
 ### "replaceROWS" method to make "[" and "[<-" work out-of-the-box.
@@ -441,6 +483,11 @@ setMethod("normalizeSingleBracketReplacementValue", "ANY",
 ### See "extractROWS" and "replaceROWS" methods for Hits objects for an
 ### example.
 ###
+### mergeROWS() is a composition of replaceROWS() and bindROWS() to
+### support appending in [<-(). Vector subclasses never need to
+### implement mergeROWS(), but a custom method may be useful for
+### e.g. optimization.
+###
 
 setGeneric("extractROWS", signature=c("x", "i"),
     function(x, i) standardGeneric("extractROWS")
@@ -450,13 +497,25 @@ setGeneric("replaceROWS", signature=c("x", "i"),
     function(x, i, value) standardGeneric("replaceROWS")
 )
 
+setGeneric("mergeROWS", signature=c("x", "i"),
+    function(x, i, value) standardGeneric("mergeROWS")
+)
+
+setGeneric("extractCOLS", signature=c("x", "i"),
+    function(x, i) standardGeneric("extractCOLS")
+)
+
+setGeneric("replaceCOLS", signature=c("x", "i"),
+    function(x, i, value) standardGeneric("replaceCOLS")
+)
+
 default_extractROWS <- function(x, i)
 {
   if (is.null(x) || missing(i))
     return(x)
   ## dynamically call [i,,,..,drop=FALSE] with as many "," as length(dim)-1
   ndim <- max(length(dim(x)), 1L)
-  i <- normalizeSingleBracketSubscript(i, x, allow.NAs=TRUE)
+  i <- normalizeSingleBracketSubscript(i, x, allow.NAs=TRUE, allow.append=TRUE)
   args <- rep.int(list(quote(expr=)), ndim)
   args[[1]] <- i
   args <- c(list(x), args, list(drop=FALSE))
@@ -464,6 +523,11 @@ default_extractROWS <- function(x, i)
 }
 
 default_replaceROWS <- function(x, i, value)
+{
+    mergeROWS(x, i, value)
+}
+
+default_mergeROWS <- function(x, i, value)
 {
   if (is.null(x))
     return(x)
@@ -549,6 +613,7 @@ setMethod("[", "LLint", subset_along_ROWS)
 
 setMethod("replaceROWS", c("ANY", "ANY"), default_replaceROWS)
 
+setMethod("mergeROWS", c("ANY", "ANY"), default_mergeROWS)
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### normalizeDoubleBracketSubscript()
